from enum import Enum
from nmigen import *
from nmigen.hdl.rec import *
from nmigen.asserts import *
from nmigen.cli import main
from bitarray import *
from bitarray.util import *


# WARNING!
# Signal(range(8)) will generate a THREE(3) wide signal, not enough to store 8
# we need to check all our widths to verify we haven't fallen prey to this


class ArbitraryWidthMemoryLayout(Layout):
    def __init__(self, *, data_width, address_width): # the * forces keyword args
        super().__init__([
            # DATA
            ("r_data",        unsigned(data_width)),      # OUTPUT

            # ADDRESS
            ("r_addr",        unsigned(address_width)),   # INPUT

            # FLOW CONTROL UPSTREAM
            ("valid_in",       1),                        # INPUT
            ("ready_out",      1),                        # OUTPUT

            # FLOW CONTROL DOWNSTREAM
            ("valid_out",      1),                        # OUTPUT
            ("ready_in",       1),                        # INPUT

        ])

class ArbitraryWidthMemoryBus(Record):
    def __init__(self, *, data_width, address_width):
        super().__init__(ArbitraryWidthMemoryLayout(data_width=data_width, address_width=address_width))



def folder(unwrapped_bits, *, desired_width):
    surplus = len(unwrapped_bits) % desired_width

    if (surplus != 0):
        padding = desired_width - surplus
        print("Padding", padding)
        unwrapped_bits.extend(zeros(padding))

    assert(len(unwrapped_bits) % desired_width == 0)

    rewrapped_array = []

    for idx in range(len(unwrapped_bits)//desired_width):
        #print("slicer",               unwrapped_bits[(idx*desired_width):(idx*desired_width+desired_width)])
        rewrapped_array.append(ba2int(unwrapped_bits[(idx*desired_width):(idx*desired_width+desired_width)]))

    return rewrapped_array

def refolder(fake_array, *, fake_width, desired_width):
    unwrapped_bits = bitarray("")

    # unwrap the bits
    for item in fake_array:
        tmp = int2ba(item, length=fake_width)
        unwrapped_bits.extend(tmp)

    # then pad it

    surplus = len(unwrapped_bits) % desired_width

    if (surplus != 0):
        padding = desired_width - surplus
        print("Padding", padding)
        unwrapped_bits.extend(zeros(padding))

    assert(len(unwrapped_bits) % desired_width == 0)

    rewrapped_array = []

    for idx in range(len(unwrapped_bits)//desired_width):
        print("slicer",               unwrapped_bits[(idx*desired_width):(idx*desired_width+desired_width)])
        rewrapped_array.append(ba2int(unwrapped_bits[(idx*desired_width):(idx*desired_width+desired_width)]))

    return rewrapped_array



class ArbitraryWidthMemory(Elaboratable):
        def __init__(self, *, fake_data_width, fake_address_width, backing_memory_data_width, backing_memory_address_width, initial_data):
            self.fake_data_width    = fake_data_width
            self.fake_address_width = fake_address_width

            def is_power_of_two(n):
                return (n & (n-1) == 0)

            assert(is_power_of_two(backing_memory_data_width))

            self.backing_memory_data_width      = backing_memory_data_width
            self.backing_memory_data_width_bits = backing_memory_data_width.bit_length() - 1
            assert(2**self.backing_memory_data_width_bits == backing_memory_data_width)

            self.backing_memory_address_width = backing_memory_address_width

            self.backing_memory_length = 256 #len(initial_data)

            assert(2**backing_memory_address_width >= self.backing_memory_length)

            self.bus = ArbitraryWidthMemoryBus(data_width=fake_data_width, address_width=fake_address_width)
            self.backing_memory = Memory(width=backing_memory_data_width, depth=self.backing_memory_length, init=initial_data)

        def elaborate(self, platform):
            m = Module()

            bus = self.bus

            # regs to persist output of last read onto the output bus until it's accepted by the downstream consumer
            last_r_data       = Signal(self.fake_data_width)
            output_txn_not_happened = Signal(1)



            unwrapped_bit_index = Signal(range(self.backing_memory_length*self.backing_memory_data_width)) #probably cannot be made shorter.

            LS_bit_index = Signal(self.backing_memory_data_width)
            LS_bit_index_internal = Signal(self.backing_memory_data_width)
            MS_bit_index = Signal(self.backing_memory_data_width)
            end_bit_pseudo_index = Signal(range(self.backing_memory_length*self.backing_memory_data_width)) # can be made shorter
            additional_words = Signal(self.backing_memory_address_width) # can also be amde shorter

            starting_word = Signal(self.backing_memory_address_width)



            # State across clock cycles
            next_address = Signal(self.backing_memory_address_width)
            additional_words_regd = Signal(self.backing_memory_address_width)
            end_bit_pseudo_index_regd = Signal(range(self.backing_memory_length*self.backing_memory_data_width)) # can be made shorter

            # pseudo-output signals, we connect these to memory and a shift register later:
            valid_fetch = Signal(1)
            fetch_address = Signal(self.backing_memory_address_width)


            m.submodules.read_port  = read_port  = self.backing_memory.read_port()
            m.d.comb += read_port.addr.eq(fetch_address)


            shreg = Signal(self.fake_data_width)
            current_slice = Signal(self.fake_data_width)
            lower_bits_cut = Signal(self.backing_memory_data_width)
            top_cut = Signal(range(self.backing_memory_data_width+1))

            valid_shreg_bits = Signal(range(self.fake_data_width))
            shreg_new_bits = Signal(range(self.backing_memory_data_width+1))

            lingering_txn = Signal(1)

            # This is the slicer; it takes the word from the memory read port and only keeps
            # bits [MS_bit_index inclusive, LS_bit_index inclusive] from it. If we want a whole
            # byte, for example, we set MS_bit_index=7, LS_bit_index=0.
            # If we want the bit above the least-significant bit, we set MS_bit_index=1=LS_bit_index

            # Note that this leaves the top bits zero, which is excellent because it means that
            # we can shift it into the shift register without any trouble.

            m.d.comb += lower_bits_cut.eq(read_port.data>>LS_bit_index)
            m.d.comb += top_cut.eq(self.backing_memory_data_width-MS_bit_index-1+LS_bit_index)
            m.d.comb += current_slice.eq(((lower_bits_cut<<top_cut)&0xff)>>top_cut)
            #                                                        ^ FIXME, this needs to be generated at synthesis time and needs to be all-ones as wide as the memory word!

            # Here we calculate the number of bits in the sliced word, which determines the
            # shift that needs to be done on the shift register:
            m.d.comb += shreg_new_bits.eq(MS_bit_index-LS_bit_index+1)

            with m.FSM() as fsm:
                with m.State("RESET"):
                    m.next ="SINGLE"

                with m.State("SINGLE"):
                    m.d.comb += bus.ready_out.eq(1)


                    with m.If(bus.valid_in == 1):
                        # the memory address and bit-index computation goes as follows:

                        # We calculate an unwrapped bit index by multiplying the fake index by the fake data width
                        m.d.comb += unwrapped_bit_index.eq(bus.r_addr * self.fake_data_width)

                        # The memory address is simply the unwrapped bit index divided by the memory data width
                        # since the memory data width is a power of two, this is a shift!
                        m.d.comb += starting_word.eq(unwrapped_bit_index[self.backing_memory_data_width_bits:])

                        # We say we'll do a fetch at that address:
                        m.d.comb += fetch_address.eq(starting_word)

                        # We start our cut at the unwrapped bit index modulo the memory data width.
                        # since the memory data width is a power of two, this is the K-least-significant-bits
                        # of the unwrapped bit index
                        m.d.comb += LS_bit_index_internal.eq(unwrapped_bit_index[:self.backing_memory_data_width_bits])
                        m.d.sync += LS_bit_index.eq(         unwrapped_bit_index[:self.backing_memory_data_width_bits])


                        # Here's where they start trying to trick you. We need to handle the case where the end of the
                        # fake word goes beyond a real memory word.
                        m.d.comb += end_bit_pseudo_index.eq(LS_bit_index_internal + self.fake_data_width - 1)

                        # So here we determine if there's any need for additional memory words:
                        m.d.comb += additional_words.eq(end_bit_pseudo_index[self.backing_memory_data_width_bits:])

                        with m.If(additional_words == 0):
                            # No additional words, calculate which bits we need from the sole word we're fetching:
                            m.d.sync += MS_bit_index.eq(end_bit_pseudo_index[:self.backing_memory_data_width_bits])
                            m.d.comb += self.bus.r_data.eq(current_slice)

                        with m.Else():
                            # Additional words needed, so we fetch the entire remaining part of this word
                            m.d.sync += MS_bit_index.eq(self.backing_memory_data_width-1)
                            # and we register state for the next cycle[s]
                            m.d.sync += next_address.eq(fetch_address + 1)
                            # rather than keeping track of the next address and the final address,
                            # we keep track of the current address and how many additional words are left,
                            # maybe that saves a few LUTs by avoiding wide compares, who knows.
                            m.d.sync += additional_words_regd.eq(additional_words)
                            m.d.sync += end_bit_pseudo_index_regd.eq(end_bit_pseudo_index)

                            m.d.sync += shreg.eq(current_slice)


                            m.next="MULTIPLE"


                with m.State("MULTIPLE"):
                    m.d.comb += bus.ready_out.eq(0)
                    m.d.sync += shreg.eq(shreg << (shreg_new_bits) | current_slice)


                    with m.If(additional_words_regd == 1):
                        # We handle the final (potentially partial word) fetch here

                        m.d.sync += LS_bit_index.eq(0)

                        # But this is the last word, so we may not have to include the whole word!
                        m.d.sync += MS_bit_index.eq(end_bit_pseudo_index_regd[:self.backing_memory_data_width_bits])
                        m.d.comb += fetch_address.eq(next_address)

                        m.d.sync += lingering_txn.eq(0)
                        m.next = "FINALIZE"
                    with m.Else():
                        # We handle the non-special case here, so fetch the whole word
                        m.d.sync += LS_bit_index.eq(0)
                        m.d.sync += MS_bit_index.eq(self.backing_memory_data_width-1)
                        # and increment the address and decrement the remaining words counter

                        m.d.sync += next_address.eq(next_address + 1)
                        m.d.sync += additional_words_regd.eq(additional_words_regd - 1)

                        m.d.comb += fetch_address.eq(next_address)

                with m.State("FINALIZE"):
                    # final word fetch has happened, we now slice it and present it on the downstream data interface
                    m.d.comb += bus.valid_out.eq(1)

                    with m.If(lingering_txn == 0):
                        m.d.comb += bus.r_data.eq(     shreg << (shreg_new_bits) | current_slice)
                        m.d.sync += last_r_data.eq(shreg << (shreg_new_bits) | current_slice)
                        m.d.sync += lingering_txn.eq(1)
                    with m.If(lingering_txn == 1):
                        m.d.comb += bus.r_data.eq(last_r_data)

                    with m.If(bus.ready_in == 1):
                        m.next = "SINGLE"


            return m



# This is non-synthesizable but is intended to provide a model for formal verification.
# We initialize the "sim_memory" with the same data that's in the 


class GoldenArbitraryWidthMemory(Elaboratable):
        def __init__(self, *, data_width, address_width, sim_memory):
            self.data_width = data_width
            self.address_width = address_width
            self.sim_memory_size = len(sim_memory)
            self.fake_memory = Memory(width=data_width, depth=self.sim_memory_size, init=sim_memory)

            self.bus = ArbitraryWidthMemoryBus(data_width=data_width, address_width=address_width)

        def elaborate(self, platform):
            m = Module()

            write_ptr = Signal(range(self.sim_memory_size))

            m.submodules.read_port  = read_port  = self.fake_memory.read_port()
            bus = self.bus

            with m.If(bus.valid_in == 1):
                #m.d.comb += read_port.en.eq(0)
                m.d.comb += read_port.addr.eq(bus.r_addr)
                m.d.comb += bus.r_data.eq(read_port.data)
                m.d.sync += bus.valid_out.eq(1)


            return m







class DummyPlug(Elaboratable):

    def elaborate(self, platform):
        m = Module()



        fake_data_width=4
        fake_address_width=8

        backing_memory_data_width=8
        backing_memory_address_width=8

        fake_bits = fake_data_width * (2** fake_address_width)

        real_bits = backing_memory_data_width * (2** backing_memory_address_width)

        assert(fake_bits <= real_bits)

        unwrapped_bits = urandom(fake_bits)
        fake_array = folder(unwrapped_bits, desired_width=fake_data_width)

        backing_array = folder(unwrapped_bits, desired_width=backing_memory_data_width)

        m.submodules.AWMem = AWMem = ArbitraryWidthMemory(fake_data_width=4,
                            fake_address_width=8, initial_data=backing_array,
                            backing_memory_data_width=8, backing_memory_address_width=8)

        m.submodules.FakeAWMem = FakeAWMem = GoldenArbitraryWidthMemory(data_width=fake_data_width, address_width=fake_address_width, sim_memory=fake_array)

        for k in fake_array:
            print(hex(k))
        address = Signal(fake_address_width)
        valid = Signal(1)
        ready = Signal(1)

        m.d.comb += FakeAWMem.bus.r_addr.eq(address)
        m.d.comb +=     AWMem.bus.r_addr.eq(address)

        m.d.comb += FakeAWMem.bus.valid_in.eq(valid)
        m.d.comb +=     AWMem.bus.valid_in.eq(valid)

        m.d.comb += FakeAWMem.bus.ready_in.eq(ready)
        m.d.comb +=     AWMem.bus.ready_in.eq(ready)

        counter = Signal(8, reset=0)
       # with m.If(FakeAWMem.bus.ready_out == 1):
        m.d.sync += counter.eq(counter+1)

        with m.If(counter >= 8):
            m.d.comb += ready.eq(1)

        m.d.comb += valid.eq(1)
        m.d.comb += address.eq(counter)


        return m



if __name__ == '__main__':
    baka =DummyPlug()
    main(baka)
    #platform.build(DummyPlug())
